import inspect

from typing import Callable, Union

import logging
from logging import handlers as logging_handlers

from .config import get_config_content
from .utils import ensure_dir
from .exceptions import InvalidOption, DuplicatedHandler, LogmeError


class LogProvider:
    """
    Get a LogmeLogger object for decorated classes and functions/methods

    *LogmeLogger object is provided as a object property of this LogProvider*
    """

    def __init__(self, decorated: Callable, config: str=None, name: str=None):
        """
        Initialization of Logger Provider, all the optional arguments should be passed from logme.log()

        Required: scope
        Optional: config, name.
                (logger configuration will resolve to the default when not specified,
                *given logme.ini file exists in root dir*)

        """

        self.decorated = decorated

        # Get the module object of the decorated object
        module_obj = inspect.getmodule(self.decorated)

        logger_name = name if name else module_obj.__name__

        config_dict = get_config_content(module_obj.__file__, name=config)

        self.logger = LogmeLogger(logger_name, config_dict)


class ModuleLogger:
    """
    Resolves to a LogmeLogger object based on the config and name passed from logme.log

    """
    def __init__(self, frame: int, config: str=None, name: str=None):
        """

        :param frame: frame number of the caller stack (1 if direct instantiate, 2 if instantiated from logme.log
        :param name: name of the logger
        :param config: configuration of the logger

        """
        module_frame = inspect.stack()[frame]
        module_obj = inspect.getmodule(module_frame.frame)

        logger_name = name if name else module_obj.__name__
        config_dict = get_config_content(module_frame.filename, name=config)

        self.logger = LogmeLogger(logger_name, config_dict)

    def __getattr__(self, attr):
        """
        Delegate the attributes and methods of self.logger to self
        """
        return getattr(self.logger, attr)


# ---------------------------------------------------------------------------
# Logger Object
# ---------------------------------------------------------------------------


class LogmeLogger:
    """
    Get a logger object with configured handlers

    """
    def __init__(self, name: str, config: dict):
        """
        :param name: name of the logger
        :param config: configuration of the logger
        """

        self.name = name
        self.config = config

        self._set_master_properties()
        self._set_handlers_from_conf()

    def __getattr__(self, attr):
        """
        Delegate all the attributes and methods of logger to LogmeLogger Object
        """
        try:
            return getattr(self.logger, attr)
        except AttributeError:
            raise AttributeError(f"LogmeLogger object has no attribute '{attr}'.")

    @property
    def logger(self):
        logger = logging.getLogger(self.name)
        logger.setLevel(self.master_level)

        return logger

    @property
    def master_formatter(self):
        return logging.Formatter(self._master_formatter, style='{')

    @master_formatter.setter
    def master_formatter(self, formatter):
        self._master_formatter = formatter
        self._set_handlers_from_conf(reconfig=True)

    @property
    def master_level(self):
        return self._get_level(self._master_level)

    @master_level.setter
    def master_level(self, level):
        self._master_level = level
        self._set_handlers_from_conf(reconfig=True)

    def _set_master_properties(self):
        master_properties = {
            '_master_formatter': self.config['formatter'],
            '_master_level': self.config['level'],
            'handler_names':  [i for i in self.config.keys() if 'Handler' in i]

        }

        for k, v in master_properties.items():
            setattr(self, k, v)

    def _set_handlers_from_conf(self, reconfig=False):
        """
        Iterate through the config dict, set the active handlers
        """
        for handler_name in self.handler_names:

            if self.config[handler_name]['active'] is False:
                continue

            level = self.config[handler_name].get('level')
            formatter = self.config[handler_name].get('formatter')

            parse_args = self._get_handler_args(handler_name)

            if reconfig:  # If true, reconfigure the existing handlers
                handler_obj = self.get_handler_by_name(handler_name)
                self._config_handler(handler_obj, level=level,
                                     formatter=formatter, set_from_master=True)
            else:
                self.add_handler(handler_name, level=level, formatter=formatter,
                                 skip_duplicate=True, **parse_args)

    def _get_handler_args(self, handler_name):
        """
        Get the args passed into handler from config
        """
        if handler_name == 'FileHandler':
            try:
                filename = self.config[handler_name]['filename']
                ensure_dir(filename)

            except (TypeError, KeyError):  # filename is None or not Passed in
                raise ValueError(f"file path for the {handler_name} must not be None")

        parse_args = {k: v for k, v in self.config[handler_name].items()
                      if k not in ['active', 'level', 'formatter']}

        return parse_args

    def _config_handler(self, handler, level: Union[str, int]=None,
                        formatter: str=None, set_from_master: bool=False):

        """
        Configure the handler's level and formatter

        :param handler: logger.Handler type object
        :param level: the level of the handler
        :param formatter: the formatter of the handler

        :param set_from_master: Set *level* or *formatter* from obj.master_level and obj.master_formatter

        """
        # Set level
        if level:
            log_level = self._get_level(level)
            handler.setLevel(log_level)
        elif set_from_master:
            handler.setLevel(self.master_level)

        # Set formatter
        if formatter:
            formatter_object = logging.Formatter(formatter, style='{')
            handler.setFormatter(formatter_object)
        elif set_from_master:
            handler.setFormatter(self.master_formatter)

    def _get_level(self, level: Union[str, int]):
        """
        Get the level number of the logger
        """
        if isinstance(level, str):
            return logging.getLevelName(level.upper())
        if isinstance(level, int):  # logging.ERROR is also type of int
            return level

    def _handler_exist(self, handler: logging.Handler) -> bool:
        """
        Check if logging handler already exists
        """
        for i in self.logger.handlers:
            handler_class = i.__class__
            handler_attr = self._get_handler_attr(handler)
            exist_attr = self._get_handler_attr(i)

            # Using type(handler) here because FileHandler inherit from StreamHandler,
            # and isinstance() will be True is checking a FileHandler object with StreamHandler class
            if type(handler) is handler_class and handler_attr == exist_attr:
                return True

        return False

    def add_handler(self, handler_type: str, formatter: str=None, level: Union[str, int]=None,
                    allow_duplicate: bool=False, skip_duplicate: bool=False, **kwargs):
        """
        Add the handler to self.logger on adhoc basis

        :param handler_type: The type of handler, e.g. StreamHandler, SocketHandler
        :param formatter: formatter to be passed to the handler
        :param level: Level for the handler
        :param allow_duplicate: *USE WITH CAUTION*, this allows duplication of handlers in the same logger
        :param skip_duplicate: Skip the duplicated handler

        :param kwargs: arguments to be passed to the handler class

        """
        try:
            handler_class = getattr(logging, handler_type)
        except AttributeError:
            handler_class = getattr(logging_handlers, handler_type)

        handler = handler_class(**kwargs)
        self._config_handler(handler, level=level,
                             formatter=formatter, set_from_master=True)

        if self._handler_exist(handler):
            if allow_duplicate:
                self.logger.addHandler(handler)
                return
            if skip_duplicate:
                return
            else:
                raise DuplicatedHandler(f"{handler_class} with the exact same configuration already exists, "
                                            f"add allow_duplicate=True to allow.")
        else:
            self.logger.addHandler(handler)

    def _get_handler_attr(self, handler: logging.Handler) -> dict:
        """
        Get the *string* attribute of the handler
        """
        handler_attr = {'formatter': handler.formatter._fmt}

        for k, v in handler.__dict__.items():
            if isinstance(v, str) or str(v).isdigit():
                handler_attr[k] = v

        return handler_attr

    def reset_config(self, config_name: str=None, config: dict=None, name: str=None):
        """
        Used for alternative config. This is normally used for module scope loggers

        :param config_name: config names specified in the config
        :param config: config dict
        :param name: logger name to be changed to
        """
        # Check if the arguments are valid
        if config and config_name:
            raise InvalidOption("Can only set keyword argument of either 'config_name' or 'config', not both.")
        if not config and not config_name:
            raise InvalidOption("must specify one of 'config_name' or 'config'.")

        caller_file_path = inspect.getframeinfo(inspect.currentframe().f_back).filename
        if config_name:
            self.config = get_config_content(caller_file_path, config_name)
        else:
            self.config = config

        if name:
            self.name = name

        # Remove existing handlers and configure new
        for i in self.logger.handlers:
            self.logger.removeHandler(i)

        self._set_master_properties()
        self._set_handlers_from_conf()

    def reconfig_handler(self, handler_name: str, level: Union[str, int]=None, formatter: str=None):
        """
        Reconfigure an existing handler's level and formatter.
        *This can be used to configure a handler for a specific logger to be different from the config in Logme.ini*

        :param handler_name: **case sensitive**. name of the handler, e.g: StreamHandler, FileHandler
        :param level: the new level to be set
        :param formatter: the new formatter to be set. '{' format

        """
        if not level and not formatter:
            raise InvalidOption("Set at least one of 'level' or 'formatter' for reconfiguration.")

        handler_obj = self.get_handler_by_name(handler_name)

        if not handler_obj:
            raise LogmeError(f"{handler_name} is not found in this logger, either use add_handle() to add this handler")

        self._config_handler(handler_obj, level=level,
                             formatter=formatter)

    def get_handler_by_name(self, handler_name: str):
        """
        Get the handler of self.logger by name
        *return None if none existent*

        :param handler_name: **case sensitive**. name of the handler, e.g: StreamHandler, FileHandler

        """
        for i in self.logger.handlers:
            if i.__class__.__name__ == handler_name:
                return i




import os
import ast
from typing import List, Union

from pathlib import Path
from contextlib import contextmanager

from .exceptions import InvalidOption, LogmeError


def conf_to_dict(conf_section: List[tuple]) -> dict:
    """
    Converting the Configparser section to a dictionary format

    :param conf_section: values from config.items('section') or config['section']
    """

    return {i[0]: conf_item_to_dict(i[1]) if '\n' in i[1] else i[1]
            for i in conf_section}


def conf_item_to_dict(parse_option: str) -> dict:
    """
    Map the configuration item(such as 'FileHandler' section) to dict.
    * Do not pass in the whole config section!*

    :param: parse_option: values from config.get('section', 'option')
    """
    try:
        str_split = parse_option.strip().split('\n')
        mapped_list = list(map(lambda x: x.split(': ', 1), str_split))

        strip_blank_recursive(mapped_list)

        return dict(mapped_list)
    except AttributeError:
        raise InvalidOption(f"option passed must be a string value, not type of '{type(parse_option).__name__}'.")
    except ValueError:
        raise InvalidOption(f"{parse_option} is not a valid option, please follow the convention of 'key: value'")


def dict_to_conf(parse_dict: dict) -> dict:
    """
    flatten nested dict for logme configuration file
    """
    flattened = {}

    for k, v in parse_dict.items():

        if isinstance(v, str):
            str_val = v
        elif isinstance(v, dict):
            val_list = [f'{k1}: {v1}' for k1, v1 in v.items()]

            str_val = '\n' + '\n'.join(val_list)
        else:
            raise ValueError(f"all values in the dict must be dict or string value, "
                             f"'{type(v).__name__}' is not allowed!")

        flattened[k] = str_val

    return flattened


def strip_blank_recursive(nested_list: list):
    """
    Strip blank space or newline characters recursively for a nested list

    *This updates the original list passed in*

    """
    if not isinstance(nested_list, list):
        raise ValueError(f"iterable passed must be type of list. not '{type(nested_list).__name__}'")

    for i, v in enumerate(nested_list):
        if isinstance(v, list):
            strip_blank_recursive(v)
        elif isinstance(v, str):
            try:
                val_ = ast.literal_eval(v.strip())
            except (ValueError, SyntaxError):  # SyntaxError raised when passing in "{asctime}::{message}"
                val_ = v.strip()

            nested_list[i] = val_


def ensure_dir(dir_path: Union[Path, str], path_type: str='parent'):
    """
    Ensure the existence of the directory,

    :param dir_path: the path to be ensured
    :param path_type: current - ensure the passed in directory exists, typically used when


    """
    path_mapping = {
        'current': Path(dir_path).resolve(),
        'parent': Path(dir_path).parent.resolve()
    }

    try:
        dir_abs = path_mapping[path_type]

        if not dir_abs.exists():
            dir_abs.mkdir(parents=True, exist_ok=True)
    except KeyError:
        raise InvalidOption(f"{path_type} is not a valid option, please pass in either 'parent' or 'current'")


def check_scope(scope: str, options: list) -> bool:
    """
    check if the scope passed is within the options.
    Used both in logme/__init__.py and providers.LogDecorator

    """
    if scope not in options:
        raise LogmeError(f"scope '{scope}' is not supported, "
                         f"please use one of {options}")

    return True


@contextmanager
def cd(dir_path: str):
    """
    Context manager for cd, change back to original directory when done
    """
    cwd = os.getcwd()
    try:
        os.chdir(os.path.expanduser(dir_path))
        yield
    finally:
        os.chdir(cwd)



class LogmeError(Exception):
    """Used as a general error for the logme module"""


class MisMatchScope(Exception):
    """Used when scope and passed callable is mismatched"""


class InvalidOption(Exception):
    """Used when the option in config file is invalid"""


class InvalidConfig(Exception):
    """Used when invalid configuration is passed"""


class DuplicatedHandler(Exception):
    """Used when an identical handler is being added on to a logger"""



from pathlib import Path
from configparser import ConfigParser, NoSectionError

from typing import Union

from .utils import conf_to_dict
from .exceptions import InvalidConfig


def get_config_content(caller_file_path: Union[str, Path], name: str=None) -> dict:
    """
    Get the config section as a dictionary

    :param caller_file_path: file path of the caller, __file__
    :param name: the name(section in logme.ini) of the config to be passed. (optional, default: 'logme')

    :return: logger configuration dictionary
    """

    init_file_path = get_ini_file_path(caller_file_path)

    config = read_config(init_file_path)

    try:
        if name:
            conf_section = config.items(name)
        else:
            conf_section = config.items('logme')

        return conf_to_dict(conf_section)
    except NoSectionError:
        raise InvalidConfig(f"'{name}' is not a valid configuration in {init_file_path}")


def read_config(file_path: Union[str, Path]) -> ConfigParser:
    """
    Read the config file given a file path,
    :param file_path: the file path which the ini file is located

    :return: ConfigParser object with section populated
    :raises: InvalidConfig, If file *does not exist* or if file is *empty*
    """
    config = ConfigParser()
    config.optionxform = str

    # Update on config.read on string format file_path to work with older version of python
    config.read(str(file_path))

    if not config.sections():
        raise InvalidConfig(f"{file_path} is not a valid config file.")

    return config


def get_ini_file_path(caller_file_path: Union[str, Path]) -> Path:
    """
    Get the logme.ini config file path

    :param caller_file_path: file path of the caller, callable.__file__

    :return: Path object of the logme.ini
    """
    conf_path = Path(caller_file_path).parent / 'logme.ini'

    if caller_file_path in [Path(Path(caller_file_path).root).resolve(),
                            Path(caller_file_path).home().resolve()]:
        raise ValueError(f"logme.ini does not exist, please use 'logme init' command in your project root.")

    if not conf_path.exists():
        return get_ini_file_path(Path(caller_file_path).parent)
    else:
        return conf_path.resolve()
